#ifndef _FUNC_H_
#define _FUNC_H_

typedef enum {
	NORMAL_RETURN=0,
	EXCEPTION_RETURN,
	TAIL_RETURN,
	NO_CONVERSION,
	STACK_UNWIND
} returntype;


typedef returntype (*rhope_func)(struct calldata *);
typedef void (*special_func) (struct object *);

#define MethodName(name,type) _f_ ## name ## _AT_ ## type

#define Func(name,numparams,callspace,localtype) \
	returntype _f_ ## name (calldata * cdata)\
	{\
		localtype *locals;\
		calldata *call;\
		returntype ret;\
		int idx;\
		if(cdata->resume)\
		{\
			locals = cdata->locals;\
			call = (calldata *)(((char *)cdata->locals) + sizeof(localtype));\
		}\
		switch(cdata->resume)\
		{\
		case 0:\
			for(idx = numparams; idx < cdata->num_params; ++idx)\
				release_ref(cdata->params[idx]); cdata->num_params = numparams;\
			cdata->num_params = numparams;\
			cdata->locals = call = alloc_cdata(cdata->ct, callspace);

#define FuncNoLocals(name,numparams,callspace) \
	returntype _f_ ## name (calldata * cdata)\
	{\
		calldata *call;\
		returntype ret;\
		int idx;\
		if(cdata->resume)\
			call = cdata->locals;\
		switch(cdata->resume)\
		{\
		case 0:\
			for(idx = numparams; idx < cdata->num_params; ++idx)\
				release_ref(cdata->params[idx]);\
			cdata->num_params = numparams;\
			cdata->locals = call = alloc_cdata(cdata->ct, callspace);

#define EndFunc	\
			return NORMAL_RETURN;\
		}\
	}

#define Method(name,type,numparams,callspace,localtype) \
	returntype MethodName(name,type) (calldata * cdata)\
	{\
		localtype *locals;\
		calldata *call;\
		returntype ret;\
		int idx;\
		if(cdata->resume)\
		{\
			locals = cdata->locals;\
			call = (calldata *)(((char *)cdata->locals) + sizeof(localtype));\
		}\
		switch(cdata->resume)\
		{\
		case 0:\
			for(idx = numparams; idx < cdata->num_params; ++idx)\
				release_ref(cdata->params[idx]); cdata->num_params = numparams;\
			cdata->num_params = numparams;\
			cdata->locals = call = alloc_cdata(cdata->ct, callspace);

#define MethodNoLocals(name,type,numparams,callspace) \
	returntype MethodName(name,type) (calldata * cdata)\
	{\
		calldata *call;\
		returntype ret;\
		int idx;\
		if(cdata->resume)\
			call = cdata->locals;\
		switch(cdata->resume)\
		{\
		case 0:\
			for(idx = numparams; idx < cdata->num_params; ++idx)\
				release_ref(cdata->params[idx]);\
			cdata->num_params = numparams;\
			cdata->locals = call = alloc_cdata(cdata->ct, callspace);
	
#define NumParams
#define CallSpace

#define ParamBase(num,var,type,convtypeid) \
	call->params[0] = cdata->params[num];\
	call->resume = 0;\
	ret = coerce_value(convtypeid, call);\
	while(ret == TAIL_RETURN)\
		ret = call->tail_func(call);\
	if(ret == EXCEPTION_RETURN)\
	{\
		for(idx = 0; idx < cdata->num_params; ++idx)\
			if(idx != num)\
				release_ref(cdata->params[idx]);\
		cdata->params[0] = call->params[0];\
		return ret;\
	}\
	cdata->params[num] = call->params[0];

#define Param(num,var,type,convtypeid) ParamBase(num,var,type,convtypeid) var = (_t_##type *)(cdata->params[num]);
#define CopiedParam(num,var,type,convtypeid) ParamBase(num,var,type,convtypeid) cdata->params[num] = copy_object(cdata->params[num]); var = (_t_##type *)(cdata->params[num]);
#define Ret(num,val) cdata->params[num] = (object *)(val);
#define Return return NORMAL_RETURN;
#define Exception
#define FuncDef(name) returntype _f_ ## name (calldata * cdata);
#define MethodDef(name,type) returntype MethodName(name,type) (calldata * cdata);

#define Call(func, numparams)\
	call->num_params = numparams;\
	call->resume = 0;\
	ret = _f_ ## func (call);\
	while(ret == TAIL_RETURN)\
		ret = call->tail_func(call);\
	if(ret == EXCEPTION_RETURN)\
	{\
		for(idx = 0; idx < cdata->num_params; ++idx)\
			release_ref(cdata->params[idx]);\
		cdata->params[0] = call->params[0];\
		return ret;\
	}
	
#define MCall(methodid, numparams)\
	call->num_params = numparams;\
	call->resume = 0;\
	ret = call_method(methodid, call);\
	while(ret == TAIL_RETURN)\
		ret = call->tail_func(call);\
	if(ret == EXCEPTION_RETURN)\
	{\
		for(idx = 0; idx < cdata->num_params; ++idx)\
			release_ref(cdata->params[idx]);\
		cdata->params[0] = call->params[0];\
		return ret;\
	}

#endif //_FUNC_H_